@@ -15,11 +15,12 @@ import (
 	"io"
 	"os"
 	"path/filepath"
-	"runtime"
+	"strconv"
 	"strings"
 	"testing"
 	"time"
 
+	"github.com/syncthing/syncthing/lib/build"
 	"github.com/syncthing/syncthing/lib/events"
 	"github.com/syncthing/syncthing/lib/fs"
 	"github.com/syncthing/syncthing/lib/ignore"
@@ -82,7 +83,7 @@ func createEmptyFileInfo(t *testing.T, name string, fs fs.Filesystem) protocol.F
 	writeFile(t, fs, name, nil)
 	fi, err := fs.Stat(name)
 	must(t, err)
-	file, err := scanner.CreateFileInfo(fi, name, fs)
+	file, err := scanner.CreateFileInfo(fi, name, fs, false)
 	must(t, err)
 	return file
 }
@@ -144,7 +145,7 @@ func TestHandleFile(t *testing.T) {
 	for _, block := range blocks[1:] {
 		found := false
 		for _, toCopyBlock := range toCopy.blocks {
-			if string(toCopyBlock.Hash) == string(block.Hash) {
+			if bytes.Equal(toCopyBlock.Hash, block.Hash) {
 				found = true
 				break
 			}
@@ -191,7 +192,7 @@ func TestHandleFileWithTemp(t *testing.T) {
 		found := false
 		block := blocks[idx]
 		for _, toCopyBlock := range toCopy.blocks {
-			if string(toCopyBlock.Hash) == string(block.Hash) {
+			if bytes.Equal(toCopyBlock.Hash, block.Hash) {
 				found = true
 				break
 			}
@@ -204,7 +205,7 @@ func TestHandleFileWithTemp(t *testing.T) {
 
 func TestCopierFinder(t *testing.T) {
 	methods := []fs.CopyRangeMethod{fs.CopyRangeMethodStandard, fs.CopyRangeMethodAllWithFallback}
-	if runtime.GOOS == "linux" {
+	if build.IsLinux {
 		methods = append(methods, fs.CopyRangeMethodSendFile)
 	}
 	for _, method := range methods {
@@ -276,7 +277,7 @@ func TestCopierFinder(t *testing.T) {
 				found := false
 				block := blocks[idx]
 				for _, pulledBlock := range pulls {
-					if string(pulledBlock.block.Hash) == string(block.Hash) {
+					if bytes.Equal(pulledBlock.block.Hash, block.Hash) {
 						found = true
 						break
 					}
@@ -284,7 +285,7 @@ func TestCopierFinder(t *testing.T) {
 				if !found {
 					t.Errorf("Did not find block %s", block.String())
 				}
-				if string(finish.file.Blocks[idx-1].Hash) != string(blocks[idx].Hash) {
+				if !bytes.Equal(finish.file.Blocks[idx-1].Hash, blocks[idx].Hash) {
 					t.Errorf("Block %d mismatch: %s != %s", idx, finish.file.Blocks[idx-1].String(), blocks[idx].String())
 				}
 			}
@@ -296,7 +297,7 @@ func TestCopierFinder(t *testing.T) {
 			}
 
 			for _, eq := range []int{2, 3, 4, 7} {
-				if string(blks[eq-1].Hash) != string(blocks[eq].Hash) {
+				if !bytes.Equal(blks[eq-1].Hash, blocks[eq].Hash) {
 					t.Errorf("Block %d mismatch: %s != %s", eq, blks[eq-1].String(), blocks[eq].String())
 				}
 			}
@@ -335,13 +336,13 @@ func TestWeakHash(t *testing.T) {
 	// both are of the same length, for example:
 	// File 1: abcdefgh
 	// File 2: xyabcdef
-	f.Seek(0, os.SEEK_SET)
+	f.Seek(0, io.SeekStart)
 	existing, err := scanner.Blocks(context.TODO(), f, protocol.MinBlockSize, size, nil, true)
 	if err != nil {
 		t.Error(err)
 	}
 
-	f.Seek(0, os.SEEK_SET)
+	f.Seek(0, io.SeekStart)
 	remainder := io.LimitReader(f, size-shift)
 	prefix := io.LimitReader(rand.Reader, shift)
 	nf := io.MultiReader(prefix, remainder)
@@ -776,7 +777,7 @@ func TestDeleteIgnorePerms(t *testing.T) {
 
 	stat, err := file.Stat()
 	must(t, err)
-	fi, err := scanner.CreateFileInfo(stat, name, ffs)
+	fi, err := scanner.CreateFileInfo(stat, name, ffs, false)
 	must(t, err)
 	ffs.Chmod(name, 0600)
 	scanChan := make(chan string, 1)
@@ -788,7 +789,7 @@ func TestCopyOwner(t *testing.T) {
 	// Verifies that owner and group are copied from the parent, for both
 	// files and directories.
 
-	if runtime.GOOS == "windows" {
+	if build.IsWindows {
 		t.Skip("copying owner not supported on Windows")
 	}
 
@@ -811,7 +812,7 @@ func TestCopyOwner(t *testing.T) {
 	// Create a parent dir with a certain owner/group.
 
 	f.mtimefs.Mkdir("foo", 0755)
-	f.mtimefs.Lchown("foo", expOwner, expGroup)
+	f.mtimefs.Lchown("foo", strconv.Itoa(expOwner), strconv.Itoa(expGroup))
 
 	dir := protocol.FileInfo{
 		Name:        "foo/bar",
@@ -985,7 +986,7 @@ func TestDeleteBehindSymlink(t *testing.T) {
 	must(t, ffs.RemoveAll(link))
 
 	if err := fs.DebugSymlinkForTestsOnly(destFs, ffs, "", link); err != nil {
-		if runtime.GOOS == "windows" {
+		if build.IsWindows {
 			// Probably we require permissions we don't have.
 			t.Skip("Need admin permissions or developer mode to run symlink test on Windows: " + err.Error())
 		} else {
@@ -1114,9 +1115,10 @@ func TestPullCaseOnlyPerformFinish(t *testing.T) {
 		t.Fatal("file is missing")
 	}
 
-	remote := *(&cur)
+	remote := cur
 	remote.Version = protocol.Vector{}.Update(device1.Short())
 	remote.Name = strings.ToUpper(cur.Name)
+
 	temp := fs.TempName(remote.Name)
 	writeFile(t, ffs, temp, contents)
 	scanChan := make(chan string, 1)
@@ -1143,7 +1145,7 @@ func TestPullCaseOnlyDir(t *testing.T) {
 }
 
 func TestPullCaseOnlySymlink(t *testing.T) {
-	if runtime.GOOS == "windows" {
+	if build.IsWindows {
 		t.Skip("symlinks not supported on windows")
 	}
 	testPullCaseOnlyDirOrSymlink(t, false)
@@ -1180,7 +1182,8 @@ func testPullCaseOnlyDirOrSymlink(t *testing.T, dir bool) {
 
 	scanChan := make(chan string, 1)
 	dbUpdateChan := make(chan dbUpdateJob, 1)
-	remote := *(&cur)
+
+	remote := cur
 	remote.Version = protocol.Vector{}.Update(device1.Short())
 	remote.Name = strings.ToUpper(cur.Name)
 
@@ -1272,7 +1275,7 @@ func TestPullCaseOnlyRename(t *testing.T) {
 }
 
 func TestPullSymlinkOverExistingWindows(t *testing.T) {
-	if runtime.GOOS != "windows" {
+	if !build.IsWindows {
 		t.Skip()
 	}
 
